#pragma once

#include <windows.h>

class SoundEmitter
{
public:
	SoundEmitter() { active = false; }
	virtual ~SoundEmitter() {}
	virtual void AddSound(int numsamples, int* samplebuffer) {}
	bool active;
};

class SoundFilter
{
public:
	SoundFilter() { }
	virtual ~SoundFilter() {}
	virtual void FilterSound(int numsamples, int* samplebuffer) {}
};


class SimpleSound : public SoundEmitter
{
public:
	static const int samplerate = 44100;
	SimpleSound(float frequency, float volume, float pan);
	virtual ~SimpleSound();
	virtual void AddSound(int numsamples, int* samplebuffer);

	int samplecount;
	// Shift values are shift per second.
	float freq;
	float freqshift;
	float vol, volshift;
	float pan, panshift;

	float wavepos;
};


class EchoFilter : public SoundFilter
{
public:
	static const int samplerate = 44100;
	static const int max_samples = 65536;
	EchoFilter();
	virtual ~EchoFilter();
	virtual void FilterSound(int numsamples, int* samplebuffer);

	void SetDelay(float delayseconds);
	void SetMix(float wetmix);

	float mix;
	float delaytime;
	int writecursor, delaysamples;

	int filterbuffer[max_samples*2]; // stereo
};


class AudioSystem
{
public:
	AudioSystem(void);
	~AudioSystem(void);

	void Start(void);
	void Stop(void);
	bool IsPlaying();

	void PlaySnd(SoundEmitter * snd);
	void AddFilter(SoundFilter * filter);

	static DWORD WINAPI AudioThreadProc(VOID* param);

	bool error;

protected:
	HANDLE hThread;
	HANDLE hExitEvent;
	HANDLE hExitWaitEvent;

	HWAVEOUT hWO;
	static const int num_buffers = 10;
	static const int buffer_samples = 512;
	static const int max_sounds = 32;
	static const int max_filters = 5;
	WAVEHDR wh[num_buffers];
	short int buffers[num_buffers][buffer_samples*2];
	int mixbuffer[buffer_samples*2];
	bool playing;

	SoundEmitter* sounds[max_sounds];
	int num_filters;
	SoundFilter* filters[max_filters];

	void ThreadFunc();

	void PlayBuffer(int buffernum);
	void RenderBuffer(int buffernum);

};
