#include "BitmapWindow.h"
#include "Bitmap.h"


extern HINSTANCE g_hInst;



BitmapWindow::BitmapWindow(int x, int y, int w, int h)
{

	WNDCLASS wc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hbrBackground=0;
	wc.hCursor=LoadCursor(NULL,MAKEINTRESOURCE(IDC_ARROW));
	wc.hIcon=0;
	wc.hInstance=g_hInst;
	wc.lpfnWndProc=BitmapWindow::WndProc;
	wc.lpszClassName="bw_class";
	wc.lpszMenuName=0;
	wc.style=0;
	RegisterClass(&wc); // This will fail on all but the first run.

	userwndproc = NULL;
	cur_x = x;
	cur_y = y;
	bmp.resize(w,h);
	bmp.cls(0);

	hWnd = CreateWindowEx(WS_EX_LAYERED|WS_EX_TOPMOST|WS_EX_TOOLWINDOW,"bw_class","Bitmap Window",0,0,0,w,h,0,0,g_hInst,this);
	SetWindowLongPtr(hWnd,GWL_USERDATA,(LONG)this);
	Redraw();
	ShowWindow(hWnd,SW_SHOW);
}
BitmapWindow::~BitmapWindow(void)
{
	DestroyWindow(hWnd);
}

void BitmapWindow::Redraw()
{
	POINT pos, src;
	SIZE dims;
	pos.x = cur_x;
	pos.y = cur_y;
	src.x=src.y=0;
	dims.cx=bmp.getWidth();
	dims.cy=bmp.getHeight();

	BLENDFUNCTION bf;
	bf.BlendOp=AC_SRC_OVER;
	bf.BlendFlags=0;
	bf.SourceConstantAlpha=255;
	bf.AlphaFormat=AC_SRC_ALPHA;

	UpdateLayeredWindow(hWnd,0,&pos,&dims,bmp.DC(),&src,0,&bf,ULW_ALPHA);
}

void BitmapWindow::Move(int x, int y)
{
	cur_x = x;
	cur_y = y;
	MoveWindow(hWnd, x, y, bmp.getWidth(), bmp.getHeight(), false);
	Redraw();
}

void BitmapWindow::SetWndProc(BmpWndProc handlerproc, void* data)
{
	userwndproc = handlerproc;
	userwndprocdata = data;
}

LRESULT WINAPI BitmapWindow::WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	BitmapWindow * bw = (BitmapWindow *)GetWindowLongPtr(hWnd,GWL_USERDATA);
	switch(msg)
	{
	case WM_CREATE:
		return 0;

	case WM_PAINT:
		if(bw)
		{
			bw->Redraw();
		}
		else
		{
			PAINTSTRUCT ps;
			BeginPaint(hWnd,&ps);
			EndPaint(hWnd,&ps);
		}
		return 0;

	case WM_NCHITTEST:
		return HTCLIENT; // ALL areas are client :) So mousemove/mousedown/mouseup always work anywhere on the window.
	}

	if(bw)
	{
		if(bw->userwndproc)
		{
			return bw->userwndproc(hWnd,msg,wParam,lParam,bw->userwndprocdata);
		}
	}
	return DefWindowProc(hWnd,msg,wParam,lParam);
}

int BitmapWindow::GetX() { return cur_x; }
int BitmapWindow::GetY() { return cur_y; }
