
#include "Caverns.h"

#include <algorithm>
#include <map>
#include <cmath>

float RandFloat()
{
	int intvalue = rand()^(rand()<<8)^(rand()<<16) & 0xFFFFFF;
	return (float)intvalue / (float)0xFFFFFF;
}
float RandGaussianFloat(int iterations)
{
	int i;
	float f = 0.0f;
	for(i=0;i<iterations;i++)
	{
		f+=RandFloat();
	}
	return f/iterations;
}

int RandInt(int bound)
{
	return ((rand()^(rand()<<8)^(rand()<<16)^(rand()<<24))&0x7FFFFFFF)%bound;
}
int RandInt()
{
	return (rand()^(rand()<<8)^(rand()<<16)^(rand()<<24));
}


Partition::Partition(int X, int Y, int W, int H)
{
	x=X; y=Y; w=W; h=H;
}
Partition Partition::SplitRandom()
{
	// Get random numbers up front so it works the same even if scale is different.
	int splitaxis;
	float split = RandGaussianFloat();
	float splitpoint = RandFloat();

	if(split*(w+h) > w) splitaxis = 1; else splitaxis = 0;

	if(w<(min_wid*2) && h<(min_hei*2)) return Partition(0,0,0,0);
	if(splitaxis==0 && (w>h*4)) splitaxis = 1;
	if(splitaxis!=0 && (h<(min_hei*2) || h>w*4)) splitaxis=0;
	Partition p;
	if(splitaxis==0 && w>=(min_wid*2))
	{ 
		// Horizontal split!
		int splitpt = (int)((w-min_wid*2)*splitpoint) + min_wid;
		if(splitpt < min_wid || (w-splitpt) < min_wid)
		{
			splitpt = splitpt;
		}
		p = Partition(x+splitpt,y,w-splitpt,h);
		w = splitpt;
	}
	else
	{
		// Vertical split!
		int splitpt = (int)((h-min_hei*2)*splitpoint) + min_hei;
		if(splitpt < min_hei || (h-splitpt) < min_hei)
		{
			splitpt = splitpt;
		}
		p = Partition(x,y+splitpt,w,h-splitpt);
		h = splitpt;
	}
	return p;
}
bool Partition::operator<(const Partition& b)
{
	return w*h < b.w*b.h;
}



int isconnected(int start1, int span1, int start2, int span2)
{

	int overlap = 0;
	if(start1<=start2 && start1+span1 > start2)
	{
		overlap = start1+span1 - start2;
		if(overlap > span2) overlap = span2;
	}
	if(start2<=start1 && start2+span2 > start1)
	{
		overlap = start2+span2 - start1;
		if(overlap > span1) overlap = span1;
	}

	if(overlap>=span1/2 || overlap >= span2/2)
	{
		// They do touch, but, is this a good connection?
		int checkpoint1, checkpoint2;
		checkpoint1 = start1+span1*3/7;
		checkpoint2 = start1+span1*4/7;
		if( checkpoint2 >= start2 && checkpoint1 <= (start2+span2))
		{ 
			checkpoint1 = start2+span2*3/7;
			checkpoint1 = start2+span2*4/7;
			if(checkpoint2 >= start1 && checkpoint1 <= (start1+span1))
			{ 
				return 2; // Yes, it's a good connection
			}
		}
		return 1; // ok connection, but not great.
	}
	return 0;
}

void Caverns::TryMakeConnection(Cavern * a, Cavern * b)
{
	if(a==b) return;
	size_t i;
	for(i=0;i<passages.size();i++)
	{
		if(passages[i].c1 == a && passages[i].c2 == b) return;
		if(passages[i].c1 == b && passages[i].c2 == a) return;	
	}
	// Not a duplicate! insert it!
	passages.push_back( Interconnect(a,b,RandInt()) );
}


Caverns::Caverns(int worldwidth, int worldheight, int approx_caverns, int seed)
{
	generate_progress = 0;
	worldw = worldwidth;
	worldh = worldheight;
	approx_num_caverns = approx_caverns;
	myseed = seed;
}

void Caverns::Generate()
{

	if(myseed!=-1) srand(myseed);
	//caverns.push_back(Cavern(worldwidth,worldheight,seed));

	// Divide world space into partitions :) (average approx_caverns)
	size_t i,j;
	size_t dividebiggest = approx_num_caverns/6+RandInt(approx_num_caverns/3);
	size_t dividerandom = approx_num_caverns/3+RandInt(approx_num_caverns*2/3);
	
	vector<Partition> partitions;

	partitions.push_back( Partition(0,0,worldw,worldh));

	for(i=0;i<dividebiggest;i++)
	{
		sort(partitions.begin(),partitions.end());
		partitions.push_back(partitions.back().SplitRandom());
	}
	for(i=0;i<dividerandom;i++)
	{
		partitions.push_back(partitions[RandInt((int)partitions.size())].SplitRandom());
	}

	// Construct caverns from (non-empty) partitions (failed split operations generate empty partitions)
	for(i=0;i<partitions.size();i++)
	{
		if(partitions[i].w != 0)
		{
			caverns.push_back( Cavern(partitions[i].x, partitions[i].y, partitions[i].w, partitions[i].h, RandInt(), -1) );
		}
	}

	// Work out the neighboring caverns (use to build connections - only allow connections between neighboring caverns)
	map<size_t,vector<size_t> > connections;
	map<size_t,vector<size_t> > backupconnections;
	for(i=0;i<caverns.size();i++)
	{
		for(j=0;j<caverns.size();j++)
		{
			int n;
			if(i>=j) continue;

			// does top of i touch bottom of j?
			if(caverns[i].y == (caverns[j].y + caverns[j].h)) {
				if((n=isconnected(caverns[i].x,caverns[i].w,caverns[j].x,caverns[j].w))>0) {
					if(n==2) { connections[i].push_back(j); connections[j].push_back(i); }
					else { backupconnections[i].push_back(j); backupconnections[j].push_back(i); }
					continue;
				}
			}
			// does bottom of i touch top of j?
			if((caverns[i].y + caverns[i].h) == caverns[j].y) {
				if((n=isconnected(caverns[i].x,caverns[i].w,caverns[j].x,caverns[j].w))>0) {
					if(n==2) { connections[i].push_back(j); connections[j].push_back(i); }
					else { backupconnections[i].push_back(j); backupconnections[j].push_back(i); }
					continue;
				}
			}
			// does left of i touch right of j?
			if(caverns[i].x == (caverns[j].x + caverns[j].w)) {
				if((n=isconnected(caverns[i].y,caverns[i].h,caverns[j].y,caverns[j].h))>0) {
					if(n==2) { connections[i].push_back(j); connections[j].push_back(i); }
					else { backupconnections[i].push_back(j); backupconnections[j].push_back(i); }
				}
			}
			// does right of i touch left of j?
			if((caverns[i].x + caverns[i].w) == caverns[j].x) {
				if((n=isconnected(caverns[i].y,caverns[i].h,caverns[j].y,caverns[j].h))>0) {
					if(n==2) { connections[i].push_back(j); connections[j].push_back(i); }
					else { backupconnections[i].push_back(j); backupconnections[j].push_back(i); }
				}
			}
		}
	}

	// If any caverns don't have connections, install backup connections.
	for(i=0;i<caverns.size();i++)
	{
		if(connections[i].size() == 0)
		{
			for(j=0;j<backupconnections[i].size();j++)
			{
				connections[i].push_back(backupconnections[i][j]); 
				connections[backupconnections[i][j]].push_back(i);
			}
		}
	}


	// Make cavern connections
	const size_t additional_connections = 15;
	vector<bool> connected;
	vector<size_t> unconnected;
	vector<size_t> trueconnections;

	connected.resize(caverns.size());

	// find "Base" cavern
	for(i=0;i<caverns.size();i++)
	{
		if(caverns[i].x==0 && caverns[i].y == 0)
		{
			base_cavern = (int)i;
		}
		else
		{
			unconnected.push_back(i);
		}
	}
	// We will eventually put the entrance in the base cavern, but not yet.
	connected[base_cavern] = true;



	while(unconnected.size() > 0)
	{ // Randomly connect all unconnected caverns first
		i = RandInt(unconnected.size());
		trueconnections.clear();
		for(j=0;j<connections[unconnected[i]].size();j++)
		{
			if(connected[connections[unconnected[i]][j]])
			{
				trueconnections.push_back(connections[unconnected[i]][j]);
			}
		}
		if(trueconnections.size() > 0)
		{
			j = trueconnections[RandInt(trueconnections.size())];
			// This will always succeed, in this context.
			TryMakeConnection(&caverns[unconnected[i]],&caverns[j]);
			connected[unconnected[i]] = true;
			unconnected.erase(unconnected.begin()+i);
		}
	}

	// Randomly add additional connections second
	for(i=0;i<additional_connections;i++)
	{
		j = RandInt(caverns.size());
		TryMakeConnection(&caverns[j],&caverns[connections[j][RandInt(connections[j].size())]]);
	}


	generate_progress = 32;
	const int caverns_progress = 96; // Caverns have 96 progress elements to play with.
	// Build caverns (after all random generation is done)
	for(i=0;i<caverns.size();i++)
	{
		caverns[i].Generate();
		generate_progress = 32 + (i+1)*caverns_progress/caverns.size();
	}
	
	generate_progress = 32+caverns_progress; // enforce progress level.

	const int passages_progress = 120; // Passages have 120 progress elements to play with.
	// Build connections (once the caverns are ready to go)
	for(i=0;i<passages.size();i++)
	{
		passages[i].Generate(this);
		generate_progress = 32 + caverns_progress + (i+1)*passages_progress/passages.size();
	}
	generate_progress = 248; // Almost Complete!

	// Generate "goal" - top-left-most cavern will have an exit.

	j = base_cavern;

	{
		
		float dx,dy,l,p;
		dx = caverns[j].w/2.0f;
		dy = caverns[j].h/2.0f;
		l = sqrt(dx*dx+dy*dy);
		entrancelength = l;
		dx = dx / l;
		dy = dy / l;

		float x,y;
		for(p=0;p<l;p+=5)
		{
			x = dx*p;
			y = dy*p;

			caverns[j].bmp.filledcircle((int)x,(int)y,(int)( entranceradius) ,Color(0));
		}

		entrancebmp.resize(entranceradius,entranceradius);
		entrancebmp.cls(0);
		for(i=0;i<entranceradius;i+=5)
		{
			int n = 255-(i*255/entranceradius);
			Color c = Color(n,n,255);
			entrancebmp.filledcircle(0,0-i,entranceradius,c);
		}
		if(entrancelength<400) entrancelength=400;
	}






	generate_progress = 256; // Complete!
}

void Caverns::Render(Bitmap & bmp, int worldx, int worldy)
{
	int bw,bh;
	size_t i;
	bw = bmp.getWidth();
	bh = bmp.getHeight();

	// Failsafe :P to not get rendering glitches at the entrance
	bmp.Blit(caverns[base_cavern].bmpbg,-worldx,-worldy);

	// Render caverns!
	for(i=0;i<caverns.size();i++)
	{
		caverns[i].Render(bmp,worldx,worldy);
	}

	// Render passageways that connect caverns
	for(i=0;i<passages.size();i++)
	{
		passages[i].Render(bmp,worldx,worldy);
	}

	// Render other world objects?

	// Render entrance
	if(worldx < entrancebmp.getWidth() && worldy < entrancebmp.getHeight())
	{
		bmp.Blit(entrancebmp,-worldx,-worldy);
	}
}

bool Caverns::IsSolid(int x, int y)
{
	size_t i;
	// Check passageways for owner
	for(i=0;i<passages.size();i++)
	{
		if(passages[i].Contains(x,y))
		{
			return passages[i].IsSolid(x,y);
		}
	}

	// Check caverns for owner
	for(i=0;i<caverns.size();i++)
	{
		if(caverns[i].Contains(x,y))
		{
			return caverns[i].IsSolid(x,y);
		}
	}
	return true;
}
double Caverns::CollideRay(float x, float y, float dx, float dy)
{
	double length = 0.0;
	length = sqrt(dx*dx+dy*dy);
	dx = (float)(dx/length);
	dy = (float)(dy/length);
	length = 0.0;
	while(length < 1200.0)
	{
		if(IsSolid((int)(x+dx*length), (int)(y+dy*length))) break;
		length+=15.0;
	}
	while(IsSolid((int)(x+dx*length), (int)(y+dy*length)))
	{
		length-=1;
	}
	return length;
}
void Caverns::FindNearbyEmptyPoint(int & x, int & y, int clearanceradius)
{

}




Cavern::Cavern(int width, int height, int seed, int caverntype)
{ // This is only a test constructor;
	x=0;
	y=0;
	w = width;
	h = height;
	type = caverntype;
	myseed = seed;
	Generate();
}

Cavern::Cavern(int x_, int y_, int width, int height, int seed, int caverntype)
{
	x=x_;
	y=y_;
	w = width;
	h = height;
	myseed = seed;
	type = caverntype;
}


bool Cavern::IsSolid(int x_, int y_)
{
	return bmp.point(x_-x,y_-y).c != 0;	
}
bool Cavern::Contains(int x_, int y_)
{
	if(x_>=x && y_>=y && x_<(x+w) && y_ < (y+h)) return true;
	return false;
}

void Cavern::Render(Bitmap & dest, int worldx, int worldy)
{
	int bw,bh;
	bw = dest.getWidth();
	bh = dest.getHeight();
	if(worldx >= x+w || worldy >= y+h || worldx+bw <= x || worldy+bh <= y) return;

	// Render the bg bmp parallax (line them up when cavern is centered, but scroll at 90% of the rate otherwise.)
	int bgx, bgy;
	int bgw, bgh;
	bgw = bmpbg.getWidth();
	bgh = bmpbg.getHeight();
	bgx = bw/2 + ((x+w/2) - (worldx + bw/2))*bgw/ w - bgw/2;
	bgy = bh/2 + ((y+h/2) - (worldy + bh/2))*bgh/ h - bgh/2;
	dest.Blit(bmpbg,bgx,bgy);
	dest.Blit(bmp, x-worldx, y - worldy);
}

void Cavern::Generate()
{
	if(myseed!=-1) srand(myseed);

	int i;
	size_t j;
	// Pick type if it's up to us.


	// Setup bitmap

	mycolor = Color(180+RandInt(40),110+RandInt(40),40+RandInt(40));

	bmp.resize(w,h);
	bmp.cls(mycolor);

	int subw,subh;
	subw = w*9/10;
	subh = h*9/10;
	bmpbg.resize(subw,subh);
	Color c = Color( mycolor.red()*3/4, mycolor.green()*3/4, mycolor.blue()*3/4 );
	bmpbg.cls(c);
	int submax = subw;
	if(subh>subw) submax=subh;



	// Eat out big chunks to make a cavern shape
	int eatcount = 40+RandInt(20);
	float eatx, eaty, maxrad, rad;

	for(i=0;i<eatcount;i++)
	{
		eatx = (float)((RandGaussianFloat()) * w);
		eaty = (float)((RandGaussianFloat()) * h);
		maxrad = (float)(0.5*(w+h));
		if(maxrad > eatx) maxrad = eatx;
		if(maxrad > eaty) maxrad = eaty;
		if(maxrad > (w-eatx)) maxrad = (float)(w-eatx);
		if(maxrad > (h-eaty)) maxrad = (float)(h-eaty);
		maxrad = maxrad * 0.8f; // Don't get too close to the edges...
		rad = (RandGaussianFloat()*0.6f + 0.4f)*maxrad;
		circles.push_back(CavernCircle(eatx,eaty,rad));
		bmp.filledcircle((int)eatx,(int)eaty,(int)rad,Color(0));

	}

	// Eat out smaller versions of the cavern shape for the background.
	const int subcircles = 7;
	for(i=subcircles-1;i>=2;i--)
	{
		c = Color( mycolor.red()*i/(subcircles+(subcircles/2)), mycolor.green()*i/(subcircles+(subcircles/2)), mycolor.blue()*i/(subcircles+(subcircles/2)) );
		for(j=0;j<circles.size();j++)
		{
			float cx,cy,r;
			cx = (float)((circles[j].x - w/2) * 0.9 * i / subcircles + subw/2);
			cy = (float)((circles[j].y - h/2) * 0.9 * i / subcircles + subh/2);
			r = (float)(circles[j].rad * 0.9 * i / subcircles);

			bmpbg.filledcircle((int)cx, (int)cy, (int)r, c);
		}
	}

	// Refine the surface, simulate water movement to produce Stalactites, Stalagmites

	// If we need to, ensure the cave is navigable.

	// Texture the remaining surface


	// If we have water, add water...

	// Done!
}


Interconnect::Interconnect(Cavern * c1_, Cavern * c2_, int seed)
{
	c1 = c1_;
	c2 = c2_;
	x=y=w=h=0;
	myseed = seed;
}

void Interconnect::Render(Bitmap & dest, int worldx, int worldy)
{
	int bw,bh;
	bw = dest.getWidth();
	bh = dest.getHeight();
	if(worldx >= x+w || worldy >= y+h || worldx+bw <= x || worldy+bh <= y) return;
	
	dest.Blit(bmpbg,x-worldx,y-worldy);
	dest.Blit(bmp,x-worldx,y-worldy);

}

bool Interconnect::IsSolid(int x_, int y_)
{
	return (bmpbg.point(x_-x,y_-y).c & 0x00FFFFFF) != 0;
}
bool Interconnect::Contains(int x_, int y_)
{
	return bmpbg.point(x_-x,y_-y).c != 0;
}


void Interconnect::ClearCaveSpace()
{
	// Eat out the "empty" regions of the caverns we're attaching to.
	{
		int ovw, ovh;
		int dx,dy;
		int cdx,cdy,cw;
		int px,py;

		if(x < c1->x) { dx=(c1->x-x); cdx = 0;} else {dx = 0; cdx = x-c1->x;}
		if(y < c1->y) { dy=(c1->y-y); cdy = 0;} else {dy = 0; cdy = y-c1->y;}
		ovw = c1->w-cdx; if(w-dx<ovw) ovw=w-dx;
		ovh = c1->h-cdy; if(h-dy<ovh) ovh=h-dy;
		cw = c1->w;

		for(py=0;py<ovh;py++)
		{
			for(px=0;px<ovw;px++)
			{
				if(c1->bmp.point(px+cdx,py+cdy).c == 0)
				{
					bmp.pset(px+dx,py+dy,Color(0));
				}
			}
		}

		if(x < c2->x) { dx=(c2->x-x); cdx = 0;} else {dx = 0; cdx = x-c2->x;}
		if(y < c2->y) { dy=(c2->y-y); cdy = 0;} else {dy = 0; cdy = y-c2->y;}
		ovw = c2->w-cdx; if(w-dx<ovw) ovw=w-dx;
		ovh = c2->h-cdy; if(h-dy<ovh) ovh=h-dy;
		cw = c2->w;

		for(py=0;py<ovh;py++)
		{
			for(px=0;px<ovw;px++)
			{
				if(c2->bmp.point(px+cdx,py+cdy).c == 0)
				{
					bmp.pset(px+dx,py+dy,Color(0));
				}
			}
		}
	}
}

void Interconnect::Generate(Caverns * parent)
{
	const int tunnel_max_radius = 120;
	const int tunnel_interior_size = 80;
	if(myseed!=-1) srand(myseed);

	double c1x, c1y, c2x, c2y, dx, dy;
	double c1l, c2l, l, ls;
	c1x = c1->x+c1->w/2;
	c1y = c1->y+c1->h/2;
	c2x = c2->x+c2->w/2;
	c2y = c2->y+c2->h/2;
	dx = c2x-c1x; dy = c2y-c1y;
	l = sqrt(dx*dx+dy*dy);
	dx = dx / l; dy = dy / l;

	c1l = parent->CollideRay((float)c1x,(float)c1y,(float)dx,(float)dy);
	c2l = parent->CollideRay((float)c2x,(float)c2y,(float)-dx,(float)-dy);

	c1x += c1l * dx; c1y += c1l * dy;
	c2x -= c2l * dx; c2y -= c2l * dy;

	dx = c2x-c1x; dy = c2y-c1y;
	l = sqrt(dx*dx+dy*dy);
	dx = dx / l; dy = dy / l;

	x = (int)c1x; if(x>c2x) x=(int)c2x;
	y = (int)c1y; if(y>c2y) y=(int)c2y;
	w = (int)c1x; if(w<c2x) w=(int)c2x;
	h = (int)c1y; if(h<c2y) h=(int)c2y;
		
	x-=tunnel_max_radius;
	y-=tunnel_max_radius;
	w = (w-x)+tunnel_max_radius;
	h = (h-y)+tunnel_max_radius;

	bmp.resize(w,h);
	bmp.cls(0);


	double a,b;
	Color c;
	for(ls=0; ls<l; ls+=5)
	{
		a = (l-ls)/l;
		b = ls/l;
		c = Color( (int)(c1->mycolor.red()*a + c2->mycolor.red()*b),
					(int)(c1->mycolor.green()*a + c2->mycolor.green()*b),
					(int)(c1->mycolor.blue()*a + c2->mycolor.blue()*b));

		bmp.filledcircle((int)(c1x+dx*ls)-x,(int)(c1y+dy*ls)-y,tunnel_max_radius,c);
	}

	for(ls=0; ls<l; ls+=5)
	{
		bmp.filledcircle((int)(c1x+dx*ls)-x,(int)(c1y+dy*ls)-y,tunnel_interior_size,Color(0xFF000000));
	}

	ClearCaveSpace();
	bmpbg = bmp;


	double r,restrict;
	const int subcircles = 7;
	int i;
	for(i=subcircles-1;i>=2;i--)
	{
		if(i==subcircles)
		{
			restrict = 0;
		}
		else
		{
			restrict = tunnel_interior_size + (subcircles-i)*tunnel_interior_size/subcircles;
		}
		restrict = 0;
		
		for(ls=restrict; ls<l-restrict; ls+=5)
		{
			
			a = (l-ls)/l;
			b = ls/l;
//			c = Color( (int)(c1->mycolor.red()*a + c2->mycolor.red()*b)*i/(subcircles+(subcircles/2)),
//				(int)(c1->mycolor.green()*a + c2->mycolor.green()*b)*i/(subcircles+(subcircles/2)),
//				(int)(c1->mycolor.blue()*a + c2->mycolor.blue()*b)*i/(subcircles+(subcircles/2)));
			c = Color( (int)(c1->mycolor.red()*a + c2->mycolor.red()*b)*3/4,
				(int)(c1->mycolor.green()*a + c2->mycolor.green()*b)*3/4,
				(int)(c1->mycolor.blue()*a + c2->mycolor.blue()*b)*3/4);

			r = tunnel_interior_size * (i+1) / subcircles;
			bmp.filledcircle((int)(c1x+dx*ls)-x,(int)(c1y+dy*ls)-y,(int)r,c);
		}
		break; // only do the first layer, this looks best.
	}

	ClearCaveSpace();



	

}

