#pragma once

#include "Bitmap.h"
#include <vector>

using namespace std;

float RandFloat();
float RandGaussianFloat(int iterations = 16);
int RandInt(int bound);


class Caverns;


class CavernCircle
{
public:
	CavernCircle() { x=y=rad=0; }
	CavernCircle(float circlex, float circley, float circlerad) { x=circlex; y=circley, rad=circlerad; }
	float x,y,rad;
};


class Cavern
{
public:
	Cavern(int width, int height, int seed = -1, int caverntype = -1);
	Cavern(int x_, int y_, int width, int height, int seed, int caverntype); // doesn't generate by default

	bool IsSolid(int x_, int y_);
	bool Contains(int x_, int y_);

	void Render(Bitmap & dest, int worldx, int worldy);


	vector<CavernCircle> circles;

	int x,y;
	int w,h;
	int myseed, type;
	
	Color mycolor;

	Bitmap bmp;
	Bitmap bmpbg;

	void Generate();


protected:


};

class Interconnect
{
public:
	Interconnect(Cavern * c1_, Cavern * c2_, int seed = -1);

	void Render(Bitmap & dest, int worldx, int worldy);
	bool IsSolid(int x_, int y_);
	bool Contains(int x_, int y_);

	Cavern * c1, * c2;
	int myseed;
	int x,y,w,h;

	Bitmap bmp;
	Bitmap bmpbg;

	void ClearCaveSpace();
	void Generate(Caverns * parent);
};



class Partition
{
public:
	static const int min_wid = 400;
	static const int min_hei = 300;
	Partition() { x=y=w=h=0; }
	Partition(int X, int Y, int W, int H);
	Partition SplitRandom();
	bool operator<(const Partition& b);
	int x,y,w,h;
};

// Master class containing a bundle of caverns and interconnects
class Caverns
{
public:
	Caverns(int worldwidth, int worldheight, int approx_caverns = 30, int seed = -1);

	void Generate(); // Worker function to actually spin off and build all this stuff!


	void Render(Bitmap & bmp, int worldx, int worldy);

	bool IsSolid(int x, int y);
	double CollideRay(float x, float y, float dx, float dy);
	void FindNearbyEmptyPoint(int & x, int & y, int clearanceradius);

	void TryMakeConnection(Cavern * a, Cavern * b);

	vector<Cavern> caverns;
	vector<Interconnect> passages;

	int worldw, worldh, approx_num_caverns, myseed;

	int base_cavern;
	int generate_progress; // 0-256 :)

	static const int entranceradius = 100;
	Bitmap entrancebmp;
	float entrancelength;
};