#include <windows.h>

#include "Bitmap.h"
#include "Rasterizer.h"
#include "Transform.h"
#include "BitmapWindow.h"
#include "Caverns.h"

#include "AudioSystem.h"


#include <cmath>
#include <vector>
using namespace std; // Lazy, lazy :)

//////////////////////////////////////////////////////////////////////////
// Debug vars!
const bool debug_windows = false;
const bool debug_chars = false;
const bool debug_disable_music = false;

BitmapWindow * debug_backbmp;











//////////////////////////////////////////////////////////////////////////
// Game state, etc
HINSTANCE g_hInst;
HWND g_hWnd;
const double Pi = 3.1415926535897932384626433832795;

HMENU hMenu;
__int64 gamestarttime;
__int64 lasttime;


Bitmap bmp; // Output buffer to display
Bitmap backbmp; // Un-lit back buffer
Bitmap nextbmp; // Next frame backbuffer (used by render thread);

const int dialoguefontheight = 30;
Font dialoguefont(dialoguefontheight,"Calibri",false);

Font UIFontSmall(40,"Calibri",false);
Font UIFontBig(120,"Calibri",false);

double totalseconds; // Make these global so stuff can use them passively :)
double deltaseconds;

int basex, basey; // Position of the top left corner of the screen, in the world
int nextbasex, nextbasey; // 1-frame lag due to split rendering, this is the coordinates for the next screen.

AudioSystem audio;
//EchoFilter echoes;


Caverns *c = NULL;
bool CavernCreated = false;
bool LimboFrame = true; // The first frame rendered will be in limbo due to the 1-frame lag, so special case it.

bool WonGame = false;


HANDLE hRenderThread;
HANDLE hMusicThread;

HANDLE hRenderExitThread;
HANDLE hRenderExitedThread;
HANDLE hRenderCreateLevel;
HANDLE hRenderCreatedLevel;
HANDLE hRenderDrawLevel;
HANDLE hRenderDrewLevel;
HANDLE hRenderLockCavern;
HANDLE hRenderLockedCavern;
HANDLE hRenderUnlockCavern;

HANDLE hMusicExitThread;
HANDLE hMusicExitedThread;

















void SetFullAlpha(Bitmap& bmp)
{
	int i,n;
	unsigned long * bits;
	n = bmp.getWidth()*bmp.getHeight();
	bits = bmp.getBits();
	for(i=0;i<n;i++) bits[i] |= 0xFF000000;
}







class Ghost
{
public:
	static const int num_sides = 20;
	static const size_t frame_lag = 5;

	Ghost() 
	{ // Set default values.
		worldx = 100;
		worldy = 100;
		r=1;g=1;b=0.5f;
		or=1;og=1;ob=1;oa=0.5;

		int i;
		for(i=0;i<num_sides;i++)
		{
			corner_start[i] = 0;
			corner_period[i] = 0;
		}
	}
	float worldx, worldy;
	double corner_start[num_sides]; // used for effect, not written yet :)
	double corner_period[num_sides];
	float r,g,b; // "ghost" color
	float or,og,ob,oa; // outer color
	vector<float> lagworldx, lagworldy;

	void Render(Bitmap& bmp)
	{
		const float innerradius = 20.0f;
		const float outerradius = 70.0f;
		const float variance = 8.0f;

		const double minperiod = 3;
		const double maxperiod = 10;

		double time = totalseconds;
		vector<VertexColor> colorvtxs;

		colorvtxs.clear();
		colorvtxs.reserve(num_sides*3*(2+frame_lag));
		
		int i;

		float x = worldx - (float)basex;
		float y = worldy - (float)basey;

		float lagx = 0.0f;
		float lagy = 0.0f;

		lagx=worldx;
		lagy=worldy;
		const float max_lag_distance = 7.0f;
		const float lag_distance_step = 2.0f;
		for(i=(int)lagworldx.size();i>0;)
		{ // Enforce a distance restriction
			float dx,dy,l,lag_allowed;
			i--;
			dx = lagworldx[i] - lagx;
			dy = lagworldy[i] - lagy;
			l = sqrt(dx*dx+dy*dy);
			lag_allowed = max_lag_distance + lag_distance_step*i;
			if(l>lag_allowed)
			{
				lagworldx[i] = lagx + dx*lag_allowed/l;
				lagworldy[i] = lagy + dy*lag_allowed/l;
			}
			lagx = lagworldx[i];
			lagy = lagworldy[i];
		}



		float angle, lastangle, sn, cs, lsn, lcs,rad;
		lastangle = 0;
		float corner_loc[num_sides];
		// Update outer shimmery bit
		for(i=0;i<num_sides;i++)
		{
			if(time > corner_start[i] + corner_period[i])
			{
				corner_start[i] = time;
				corner_period[i] = RandFloat()*(maxperiod-minperiod) + minperiod;
			}
			corner_loc[i] = (float)sin(Pi*2*(time-corner_start[i])/corner_period[i]);
		}

		lastangle = (float)(time/4);
		rad = outerradius;
		rad = outerradius + (float)(corner_loc[num_sides-1]*variance);
		for(i=0;i<num_sides;i++) // Draw outer shimmery bit
		{
			angle = (float)(lastangle + 2*Pi/num_sides);
			lsn = sin(lastangle)*rad; lcs = cos(lastangle)*rad;
			rad = outerradius + (float)(corner_loc[i]*variance);
			sn = sin(angle)*rad; cs = cos(angle)*rad;

			colorvtxs.push_back(VertexColor(x,y,1,or,og,ob,oa));
			colorvtxs.push_back(VertexColor(x+sn,y+cs,1,or,og,ob,0));
			colorvtxs.push_back(VertexColor(x+lsn,y+lcs,1,or,og,ob,0));

			lastangle = angle;
		}

		lastangle = 0;
		rad = innerradius;
		size_t lag;
		float lagmod;
		for(lag=0;lag<=frame_lag;lag++)
		{
			if(lag < lagworldx.size())
			{
				lagx = lagworldx[lag] - (float)basex;
				lagy = lagworldy[lag] - (float)basey;
			}
			else
			{
				lagx = x; lagy = y;
			}
			lagmod = (lag+1.0f)/(frame_lag+1.0f);
			
			for(i=0;i<num_sides;i++) // Draw inner core thing.
			{
				angle = (float)(lastangle + 2*Pi/num_sides);
				lsn = sin(lastangle)*rad*lagmod; lcs = cos(lastangle)*rad*lagmod;
				sn = sin(angle)*rad*lagmod; cs = cos(angle)*rad*lagmod;

				colorvtxs.push_back(VertexColor(lagx,lagy,1,r,g,b,1.0f*lagmod));
				colorvtxs.push_back(VertexColor(lagx+sn,lagy+cs,1,r,g,b,0.5f*lagmod));
				colorvtxs.push_back(VertexColor(lagx+lsn,lagy+lcs,1,r,g,b,0.5f*lagmod));

				lastangle = angle;
			}

		}


		Rasterizer<false,true> ras;
		ras.Bind(&bmp);

		ras.RasterizeTriangleList<VertexColor,PixelShaderColor,BlendModeAlpha>(colorvtxs);

		lagworldx.push_back(worldx);
		lagworldy.push_back(worldy);
		if(lagworldx.size() > frame_lag) lagworldx.erase(lagworldx.begin());
		if(lagworldy.size() > frame_lag) lagworldy.erase(lagworldy.begin());
	}
};







template<typename VtxFormat>
void PixelShaderLightBubble(VtxFormat & input, PixelColor & output, std::vector<Bitmap*>& textures)
{
	ReadPx(textures[0]->point((int)(input.x-0.5f),(int)(input.y-0.5f)).c,output); // Copy the pixel from the other texture...
	output.a = input.a; // but replace alpha channel with ours.
}


void DrawLightBubble(Bitmap & bmp, float worldx, float worldy, float nearrad, float farrad)
{
	const int num_sides = 30;
	vector<VertexColor> colorvtxs;

	float x = worldx - (float)basex;
	float y = worldy - (float)basey;

	colorvtxs.clear();
	colorvtxs.reserve(num_sides*15);
	float angle, lastangle, sn, cs, lsn, lcs, rad, rad_far,rad_mid, alpha_mid;
	lastangle = 0;
	int i;

	float radnear[num_sides];
	float radmid[num_sides];
	float radfar[num_sides];
	float alphamid[num_sides];
	lastangle = 0;
	for(i=0;i<num_sides;i++)
	{	// Figure out how far we can cast the light rays
		// (TODO: for now we just assume max radius)
		angle = (float)(lastangle + 2*Pi/num_sides); // 0 = up, and then clockwise.
		sn = sin(angle); cs = cos(angle);
		radnear[i] = nearrad;
		radmid[i] = farrad;
		radfar[i] = farrad;
		alphamid[i] = 0;
		double dist = c->CollideRay(worldx,worldy,sn,cs);
		if(dist < nearrad)
		{
			radnear[i] = (float)dist;
			radmid[i] = (float)(dist+40);
			radfar[i] = radmid[i];
		}
		else if(dist < farrad)
		{
			radmid[i] = (float)(dist);
			alphamid[i] = (farrad-radmid[i]) / (farrad-nearrad);
			if((dist+40)<farrad) radfar[i] = (float)(dist+40);
		}
		lastangle = angle;
	}


	lastangle = 0;
	rad = radnear[num_sides-1];
	for(i=0;i<num_sides;i++) // inner fully visible thingy (easy part)
	{
		angle = (float)(lastangle + 2*Pi/num_sides);
		lsn = sin(lastangle)*rad; lcs = cos(lastangle)*rad;
		rad = radnear[i];
		sn = sin(angle)*rad; cs = cos(angle)*rad;

		colorvtxs.push_back(VertexColor(x,y,1,1,1,1,1));
		colorvtxs.push_back(VertexColor(x+sn,y+cs,1,1,1,1,1));
		colorvtxs.push_back(VertexColor(x+lsn,y+lcs,1,1,1,1,1));

		lastangle = angle;
	}

	lastangle = 0;
	rad = radnear[num_sides-1];
	rad_mid = radmid[num_sides-1];
	rad_far = radfar[num_sides-1];
	alpha_mid = alphamid[num_sides-1];

	for(i=0;i<num_sides;i++) // Outer special part; 
	{
		bool render_third_tier = false;
		angle = (float)(lastangle + 2*Pi/num_sides);
		lsn = sin(lastangle); lcs = cos(lastangle);
		VertexColor a,b,c,d,bb,dd;
		if(rad_mid!=rad_far) render_third_tier = true;
		a = VertexColor(x+lsn*rad,y+lcs*rad,1,1,1,1,1);
		b = VertexColor(x+lsn*rad_mid,y+lcs*rad_mid,1,1,1,1,alpha_mid);
		bb = VertexColor(x+lsn*rad_far,y+lcs*rad_far,1,1,1,1,0);
		rad = radnear[i];
		rad_far = radfar[i];
		rad_mid = radmid[i];
		alpha_mid = alphamid[i];
		sn = sin(angle); cs = cos(angle);
		if(rad_mid!=rad_far) render_third_tier = true;
		c = VertexColor(x+sn*rad,y+cs*rad,1,1,1,1,1);
		d = VertexColor(x+sn*rad_mid,y+cs*rad_mid,1,1,1,1,alpha_mid);
		dd = VertexColor(x+sn*rad_far,y+cs*rad_far,1,1,1,1,0);

		colorvtxs.push_back(a);
		colorvtxs.push_back(b);
		colorvtxs.push_back(c);

		colorvtxs.push_back(b);
		colorvtxs.push_back(c);
		colorvtxs.push_back(d);

		if(render_third_tier)
		{
			colorvtxs.push_back(b);
			colorvtxs.push_back(bb);
			colorvtxs.push_back(d);

			colorvtxs.push_back(bb);
			colorvtxs.push_back(d);
			colorvtxs.push_back(dd);

		}
		

		lastangle = angle;
	}

	Rasterizer<false,true> ras;
	ras.Bind(&bmp);
	ras.textures.push_back(&backbmp);

	ras.RasterizeTriangleList<VertexColor,PixelShaderLightBubble,BlendModeAlpha>(colorvtxs);

}



void DrawTextBubble(Bitmap& bmp, int worldoriginx, int worldoriginy, string text)
{
	const int maxtextwidth = 380;
	const int mintextwidth = 30;
	const int padding = 10;
	const int trianglesize = 20;
	const int trianglepad = 30;

	int originx = worldoriginx-basex;
	int originy = worldoriginy-basey;
	
	Color bubblecolor = Color(255,255,255);
	Color edgecolor = Color(0,0,0);
	Color textcolor = Color(0,0,0);

	int textwidth,textheight;

	bmp.SetFont(dialoguefont);
	bmp.SetTextcolor( textcolor );
	textwidth = bmp.GetTextWidth(text.c_str());
	if(textwidth > maxtextwidth) textwidth = maxtextwidth;
	if(textwidth < mintextwidth) textwidth = mintextwidth;
	textheight = bmp.GetTextBlockHeight(text.c_str(),textwidth);


	int bubblewidth = textwidth+padding*2;
	int bubbleheight = textheight+padding*2+trianglesize;
	int bubblex,bubbley;
	bubbley = originy-bubbleheight;
	bubblex = originx-bubblewidth/2;
	// If the bubble is offscreen in this position, try to correct this situation.
	if(bubblex < 10) bubblex = 10;
	if(bubblex+bubblewidth > bmp.getWidth()-10) bubblex = bmp.getWidth()-10-bubblewidth;

	// Correct position (can't be too far away from the triangle)
	if(bubblex > originx-trianglepad) bubblex = originx-trianglepad;
	if(bubblex+bubblewidth < originx+trianglepad) bubblex = originx+trianglepad-bubblewidth;

	// Draw bubble!
	int bubblerectheight = bubbleheight-trianglesize;

	bmp.clearrect(bubblex,bubblex+bubblewidth-1,bubbley,bubbley+bubblerectheight-1,bubblecolor);
	bmp.hline(bubblex,bubblex+bubblewidth-1,bubbley,edgecolor);
	bmp.hline(bubblex,bubblex+bubblewidth-1,bubbley+bubblerectheight-1,edgecolor);
	bmp.vline(bubblex,bubbley,bubbley+bubblerectheight-1,edgecolor);
	bmp.vline(bubblex+bubblewidth-1,bubbley,bubbley+bubblerectheight-1,edgecolor);

	// Draw triangle
	int i,w;
	for(i=0;i<trianglesize;i++)
	{
		w = (7*i)/trianglesize;
		bmp.hline(originx-w,originx+w,originy-2-i,bubblecolor);
		bmp.pset(originx-w-1,originy-2-i,edgecolor);
		bmp.pset(originx+w+1,originy-2-i,edgecolor);
	}
	bmp.pset(originx,originy-1,edgecolor);

	// Draw text in bubble
	bmp.RenderTextBlock(bubblex+padding,bubbley+padding,textwidth,textheight,text.c_str());


}


float camerax, cameray;

Ghost g;
int playercavern, playerbasecavern;
//Ghost g2;

int lastrandomtext = 0;
string GreetingText()
{
	lastrandomtext += rand()+rand()+rand()+rand();
	switch(lastrandomtext%7)
	{
	case 0: return "Hi!";
	case 1: return "Hey there.";
	case 2: return "Eek! A ghost!";
	case 3: return "Haven't seen you before.";
	case 4: return "'lo";
	case 5: return "Howdy";
	case 6: return "Hi! Hi!";
	}
	return "Help! I'm covered in bugs!";
}
string RandomText()
{
	lastrandomtext += rand()+rand()+rand()+rand();
	switch(lastrandomtext%7)
	{
	case 0: return "Caverns are AWESOME!";
	case 1: return "So many caverns to explore...";
	case 2: return "Which is your favorite room?";
	case 3: return "If I were corporeal, I would totally be eating a cookie right now.";
	case 4: return "Wanna play... Oh wait, no hands.";
	case 5: return "You just missed this Jimmy kid, he just came through here.";
	case 6: return "Ah, to see the sun again...";
	}
	return "Help! I'm covered in bugs!";
}


class CavernEntity
{
public:
	CavernEntity(int owner)
	{
		owningcavern = owner;
		caverncenterx = (float)(c->caverns[owningcavern].x + c->caverns[owningcavern].w/2);
		caverncentery = (float)(c->caverns[owningcavern].y + c->caverns[owningcavern].h/2);
		float r = sqrt((float)c->caverns[owningcavern].w*c->caverns[owningcavern].w + c->caverns[owningcavern].h*c->caverns[owningcavern].h)/2;
		cavernsize = r;
		float angle = (float)(RandFloat()*2*Pi);
		r = r*RandGaussianFloat();

		float sn,cs;
		sn = sin(angle); cs = cos(angle);
		float l = (float)c->CollideRay(caverncenterx,caverncentery,cs,sn);
		if(l<r) r=l;

		g.worldx = caverncenterx + cs*r;
		g.worldy = caverncentery + sn*r;

		g.r = 0.5f + RandFloat()*0.5f;
		g.g = 0.5f + RandFloat()*0.5f;
		g.b = 0.5f + RandFloat()*0.5f;

		vx=vy=0;
		messagetime = cooltime = 0;
		greetingcooldown = 0;
	}

	void Collide()
	{
		int i;
		const int r = 30;
		for(i=10;i<=r;i+=4)
		{
			if(c->IsSolid((int)(g.worldx),(int)(g.worldy+i))) {g.worldy -= (r-i); vy *= 0.7f; }
			if(c->IsSolid((int)(g.worldx),(int)(g.worldy-i))) {g.worldy += (r-i); vy *= 0.7f; }
			if(c->IsSolid((int)(g.worldx+i),(int)(g.worldy))) {g.worldx -= (r-i); vx*=0.7f; }
			if(c->IsSolid((int)(g.worldx-i),(int)(g.worldy))) {g.worldx += (r-i); vx*=0.7f; }
		}
	}
	void Render(Bitmap & bmp)
	{
		int bw = bmp.getWidth();
		int bh = bmp.getHeight();

		// Update ghost position / velocity
		float focusstrength = 0.1f;
		float changespeed = 80.0;

		float dx, dy;
		changespeed = changespeed/(focusstrength+0.5f);
		dx = (RandGaussianFloat() - 0.5f + focusstrength*(caverncenterx-g.worldx)/cavernsize) * changespeed;
		dy = (RandGaussianFloat() - 0.5f + focusstrength*(caverncentery-g.worldy)/cavernsize) * changespeed;

		vx += (float)(dx*deltaseconds);
		vy += (float)(dy*deltaseconds);

		g.worldx += (float)(vx*deltaseconds);
		g.worldy += (float)(vy*deltaseconds);

		Collide();

		if( (g.worldx > basex - 120) && (g.worldy > basey - 120) && (g.worldx < basex + 120 + bw) && (g.worldy < basey + 120 + bh) )
		{
			DrawLightBubble(bmp,g.worldx,g.worldy,100,150);
			g.Render(bmp);
		}

	}
	void RenderUI(Bitmap & bmp)
	{
		float dx,dy,r;
		if(greetingcooldown > 0) greetingcooldown -= deltaseconds;

		if(messagetime > 0)
		{
			DrawTextBubble(bmp,(int)g.worldx,(int)g.worldy,message);
			messagetime -= deltaseconds;
		}
		else if(cooltime > 0)
		{
			cooltime -= deltaseconds;
		}
		else
		{
			dx = g.worldx - ::g.worldx;
			dy = g.worldy - ::g.worldy;
			r = sqrt(dx*dx+dy*dy);
			if(r < 150)
			{
				if(greetingcooldown > 0)
				{ // Play random message
					messagetime = 10;
					message = RandomText();					
				}
				else
				{ // Play random greeting
					messagetime = 5;
					message = GreetingText();					
				}
				greetingcooldown = 90; // 1.5 minutes - don't greet if you've been around this ghost lately.
				cooltime = 5;
			}
		}
	}
	float vx,vy;
	float caverncenterx, caverncentery, cavernsize;
	int owningcavern;
	Ghost g;

	double messagetime,cooltime, greetingcooldown;
	string message;
};

class CavernMaster
{ // Owns and operates the cavern
public:
	CavernMaster()
	{
		size_t i,j;
		vector<size_t> VacantCaverns;
		// first, find a suitable random cavern for the player: (random cavern in the bottom 1/3 of the map)
		for(i=0;i<c->caverns.size();i++)
		{
			if(c->caverns[i].y >= 4000)
			{
				VacantCaverns.push_back(i);
			}
		}
		if(VacantCaverns.size() > 0)
		{
			playerbasecavern = (int)VacantCaverns[RandInt((int)VacantCaverns.size())];
		}
		else
		{ // Oh dear. well find the deepest cavern and place player there.
			int y=0, x=0;
			for(i=0;i<c->caverns.size();i++)
			{
				if(c->caverns[i].y == y)
				{
					if(c->caverns[i].x > x)
					{
						playerbasecavern = (int)i;
						x = c->caverns[i].x;
						y = c->caverns[i].y;
					}
				}
				else if(c->caverns[i].y > y)
				{
					playerbasecavern = (int)i;
						x = c->caverns[i].x;
					y = c->caverns[i].y;
				}
			}
		}
		playercavern = playerbasecavern;


		g.worldx = (float)(c->caverns[playercavern].x + c->caverns[playercavern].w/2);
		g.worldy = (float)(c->caverns[playercavern].y + c->caverns[playercavern].h/2);

		VacantCaverns.clear();
		// Populate list of vacant caverns, out of which to spawn ghosts.
		for(i=0;i<c->caverns.size();i++)
		{
			if((int)i != playercavern && (int)i != c->base_cavern)
			{
				VacantCaverns.push_back(i);
			}
		}

		const size_t num_ghosts = 10;
		for(i=0;i<num_ghosts;i++)
		{
			j = RandInt((int)VacantCaverns.size());
			entities.push_back( CavernEntity((int)VacantCaverns[j]) );
			VacantCaverns.erase(VacantCaverns.begin()+j);
	
		}

	}


	// This is also update :D
	void Render(Bitmap & bmp)
	{
		size_t i;
		for(i=0;i<entities.size();i++)
		{
			entities[i].Render(bmp);
		}
	}

	void RenderUI(Bitmap & bmp) // Where the text bubbles happen.
	{
		size_t i;
		for(i=0;i<entities.size();i++)
		{
			entities[i].RenderUI(bmp);
		}
	}


	vector<CavernEntity> entities;

};

CavernMaster * cm = NULL;














void PlayNote(int note)
{
	// Calculate frequency!
	int MiddleA = 4*12 + 9;
	int ThisNote = note;
	int semitone_shift = ThisNote - MiddleA;
	float frequency = 440 * pow(2.0f,(float)semitone_shift/12.0f);

	audio.PlaySnd( new SimpleSound(frequency, 1.0f, 0.0f) );
}

int GetNote(int octave, int note)
{
	return octave*12+note;
}

// Music playback thread
DWORD WINAPI MusicThreadProc(void* param)
{ 
	srand(GetTickCount());

	int key = GetNote(4,0);
	int minkey = GetNote(3,4);
	int maxkey = GetNote(4,8);

	const int num_sequences = 3;
	int sequencedata[num_sequences][32] = {
		{ 0,3,7, 0,3,7, 0,3,6, 0,3,6 },
		{ 0,2,6, 0,2,6 },
		{ 0,3,8, 0,3,8 },

	};
	int sequencelengths[num_sequences] = {12, 6, 6};

	int sequenceindex = 0;
	int sequencepos = 0;
	int *sequence = sequencedata[sequenceindex];
	int sequencelength = sequencelengths[sequenceindex];

	int notetime = 800;

	while(true)
	{
		DWORD r = WaitForSingleObject(hMusicExitThread,notetime);
		if(r == WAIT_OBJECT_0)
		{ // We're done!
			ResetEvent(hMusicExitThread);
			SetEvent(hMusicExitedThread);
			return 0;
		}
		if(debug_disable_music) continue;

		if(sequence[sequencepos] != 100) // a break in the sequence.
		{
			PlayNote(key + sequence[sequencepos]);
		}
		sequencepos++;

		if(sequencepos == sequencelength)
		{
			// pick new sequence
			sequencepos = 0;
			sequenceindex = rand()%num_sequences;
			sequence = sequencedata[sequenceindex];
			sequencelength = sequencelengths[sequenceindex];

			// Randomly change music key
			switch(rand()&3)
			{
			case 1: // Big key jump
				key += rand()%9 - 4;
				if(key < minkey) key=minkey;
				if(key > maxkey) key=maxkey;
				break;
			case 2: // Raise key
				if(key != maxkey) key++;
				break;
			case 3: // Lower key
				if(key != minkey) key--;
				break;
			}
		}

	}

}






// cavern generator / rendering helper thread
DWORD WINAPI RenderThreadProc(void* param)
{

	HANDLE waitlist[5];
	while(true)
	{
		waitlist[0] = hRenderExitThread;
		waitlist[1] = hRenderCreateLevel;
		waitlist[2] = hRenderDrawLevel;
		waitlist[3] = hRenderLockCavern;

		DWORD r = WaitForMultipleObjects(4,waitlist,FALSE,INFINITE);

		if(r==(WAIT_OBJECT_0+0))
		{ // Exit!
			ResetEvent(hRenderExitThread);
			SetEvent(hRenderExitedThread);
			return 0;
		}
		else if(r==(WAIT_OBJECT_0+1))
		{ // Create a level!
			ResetEvent(hRenderCreateLevel);
			//ASSERT(c == NULL);
			CavernCreated = false;

			// Build the world...
			c = new Caverns(4000,6000,30,(int)GetTickCount());
			c->Generate();

			// And everything in it...
			cm = new CavernMaster();

			// Done!
			LimboFrame = true;
			CavernCreated = true;
			SetEvent(hRenderCreatedLevel);
		}
		if(r==(WAIT_OBJECT_0+2))
		{ // Render!
			ResetEvent(hRenderDrawLevel);
			c->Render(nextbmp,nextbasex,nextbasey);
			SetEvent(hRenderDrewLevel);
		}
		if(r==(WAIT_OBJECT_0+3))
		{
			ResetEvent(hRenderLockCavern);
			ResetEvent(hRenderUnlockCavern);
			SetEvent(hRenderLockedCavern);
			if(WAIT_OBJECT_0 == WaitForSingleObject(hRenderUnlockCavern,INFINITE))
			{
				ResetEvent(hRenderUnlockCavern);
			}
			else
			{
				// Bad state, probably going to break something...
				r = r+1; // break here
			}
		}
	}
}


void GenerateCavern()
{
	if(CavernCreated)
	{
		ResetEvent(hRenderLockedCavern);
		SetEvent(hRenderLockCavern);
		if(WAIT_OBJECT_0 == WaitForSingleObject(hRenderLockedCavern,INFINITE))
		{
			CavernCreated = false;
			delete c;
			c = NULL;
			delete cm;
			cm = NULL;
		}
		else
		{
			// Error
		}
		SetEvent(hRenderUnlockCavern);
		ResetEvent(hRenderLockedCavern);
		ResetEvent(hRenderDrewLevel);
	}
	SetEvent(hRenderCreateLevel);
	WonGame = false;
}










double movevx,movevy;
void Move(double dx, double dy)
{
	movevx = movevx*pow(0.2,deltaseconds) + dx*pow(0.8,deltaseconds);
	movevy = movevy*pow(0.2,deltaseconds) + dy*pow(0.8,deltaseconds);

	g.worldx += (float)(movevx*deltaseconds);
	g.worldy += (float)(movevy*deltaseconds);
	int i;
	const int r = 30;
	for(i=10;i<=r;i+=4)
	{
		if(c->IsSolid((int)(g.worldx),(int)(g.worldy+i))) {g.worldy -= (r-i); movevy*=0.7; }
		if(c->IsSolid((int)(g.worldx),(int)(g.worldy-i))) {g.worldy += (r-i); movevy*=0.7; }
		if(c->IsSolid((int)(g.worldx+i),(int)(g.worldy))) {g.worldx -= (r-i); movevx*=0.7; }
		if(c->IsSolid((int)(g.worldx-i),(int)(g.worldy))) {g.worldx += (r-i); movevx*=0.7; }
	}
}

void DoDebugChar(int ch)
{
	switch(ch)
	{
	case 'r': // Recreate cavern
		GenerateCavern();
		break;
	}
}

bool key_up = false, key_down = false, key_left = false, key_right = false;

void Render()
{
	__int64 curtime;

	GetSystemTimeAsFileTime((LPFILETIME)&curtime);
	deltaseconds = (curtime-lasttime)/ 10000000.0; // 100-nanosecond intervals (10^-7, seven 0s)
	totalseconds = (curtime-gamestarttime)/ 10000000.0;
	lasttime=curtime;

	srand(GetTickCount());

	bool interactiveworld = true;
	bool renderworld = true;
	int bw,bh;
	bw = bmp.getWidth();
	bh = bmp.getHeight();


	{ // show FPS?


	}

	if(WonGame)
	{
		movevx = movevy = 0;
		renderworld = false;
		bmp.cls(Color(255,255,255));
		bmp.SetFont(UIFontBig);
		bmp.SetTextcolor(Color(0,0,0));
		bmp.RenderCenteredText(bw/2, 100, "You Win!");

		bmp.SetFont(UIFontSmall);
		bmp.SetTextcolor(Color(0,0,0));
		bmp.RenderCenteredText(bw/2, 300, "(Yes, it wasn't very hard.)");
		bmp.RenderCenteredText(bw/2, 350, "(Click to play again, if you like)");

		return;
	}




	if(CavernCreated && renderworld)
	{ // Only do a lot of things if the cavern is actually created

		// Update character positions, camera
		// int basex, basey; // Position of the top left corner of the screen, in the world

		double movedx = 0,movedy = 0;
		if(interactiveworld)
		{
			if(key_up) movedy-=(float)(50);
			if(key_down) movedy+=(float)(50);
			if(key_left) movedx-=(float)(50);
			if(key_right) movedx+=(float)(50);
			Move(movedx,movedy);
		}

		camerax = g.worldx;
		cameray = g.worldy;

		if(g.worldx < c->entranceradius && g.worldy < c->entranceradius)
		{
			// Win condition!
			WonGame = true;
			// Continue this frame, but next frame will have it's own win screen.
		}

		// Interlock: Wait for render thread to complete.
		if(!LimboFrame) {
			DWORD ret = WaitForSingleObject(hRenderDrewLevel,200); // Shouldn't take a second... but if it does...
			if(ret != WAIT_OBJECT_0) 
			{
				return; // Skip a frame if it didn't complete in time. Will be even more slow if this happens...
				// Debug this if it happens, it shouldn't.
			}
			ResetEvent(hRenderDrewLevel);
		}

		basex = nextbasex;
		basey = nextbasey;
		nextbasex = (int)(camerax - bmp.getWidth()/2);
		nextbasey = (int)(cameray - bmp.getHeight()/2);
		if(nextbasex<0) nextbasex = 0;
		if(nextbasey<0) nextbasey = 0;

		// Tell render thread to get busy on the next frame :)
		backbmp = nextbmp; // Just going to glitch a little if resized; but this isn't an issue (this game won't resize)
		if(nextbmp.getWidth()!=bmp.getWidth() || nextbmp.getHeight()!=bmp.getHeight())
		{ // resize if it's the wrong size.
			nextbmp.resize(bmp.getWidth(),bmp.getHeight());
		}
		SetEvent(hRenderDrawLevel); // Tell render thread to get going on the next frame.

		if(!LimboFrame)
		{
			// Draw the illuminated areas of the world from backbmp to the front bmp
			bmp.cls(0);

			DrawLightBubble(bmp,camerax,cameray,100,300);
			//	DrawLightBubble(bmp,300,300,50,250);

			if(basex < c->entrancelength+20 || basey < c->entrancelength+20 )
			{
				DrawLightBubble(bmp,10,10,50,c->entrancelength);
			}

			// Draw the illuminated entities in the world to bmp
			cm->Render(bmp); // Rest of the world

			g.Render(bmp); // Player

			// Draw HUD/UI to bmp

			cm->RenderUI(bmp);

			//DrawTextBubble(bmp,400,280,"Hello World!");



			// Done!
		}
		else
		{
			LimboFrame = false;
		}
	}
	else
	{ // Cavern not created, or in the process of being created, or something else.
		bmp.cls(0);

		bmp.SetFont(UIFontBig);
		bmp.SetTextcolor(Color(255,255,255));

		bmp.RenderCenteredText(bw/2, 100, "Cavern Ghost");

		// Draw progress bar if we are building a cavern.
		if(!CavernCreated)
		{
			if(c != NULL)
			{
				int progress = c->generate_progress;

				bmp.SetFont(UIFontSmall);
				bmp.SetTextcolor(Color(255,255,255));

				if(progress<32)
				{
					bmp.RenderCenteredText(bw/2, bh-160, "Generating level...");
				}
				else if(progress<128)
				{
					bmp.RenderCenteredText(bw/2, bh-160, "Digging out caverns...");
				}
				else
				{	
					bmp.RenderCenteredText(bw/2, bh-160, "Digging connecting tunnels...");
				}
				 
				int rx,ry,rx2,ry2, rxprogress;
				rx = bw/4;
				rx2 = bw*3/4;
				ry = bh-100;
				ry2 = bh-50;
				rxprogress = rx + (rx2-rx)*progress/256;
				bmp.clearrect(rx-6,rx2+6,ry-6,ry2+6,Color(255,255,255));
				bmp.clearrect(rx-3,rx2+3,ry-3,ry2+3,Color(0,0,0));

				bmp.clearrect(rx,rxprogress,ry,ry2,Color(0,255,0));
			}
		}

	}

	if(debug_windows)
	{
		debug_backbmp->bmp = backbmp;
		SetFullAlpha(debug_backbmp->bmp);
		debug_backbmp->Redraw();
	}

}

void ResizeBmp()
{
	RECT rc;
	GetClientRect(g_hWnd,&rc);
	bmp.resize(rc.right-rc.left,rc.bottom-rc.top);
	backbmp.resize(rc.right-rc.left,rc.bottom-rc.top);
	InvalidateRect(g_hWnd,&rc,FALSE);
	if(debug_windows)
	{
		debug_backbmp->bmp = backbmp;
		SetFullAlpha(debug_backbmp->bmp);
		debug_backbmp->Move(debug_backbmp->GetX(), debug_backbmp->GetY());
	}
}

// This is around because it will be used for the debug windows.
bool floating_l_down;
int floating_mx, floating_my;
LRESULT WINAPI FloatingWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam, void *userdata)
{
	BitmapWindow * bmpwnd = (BitmapWindow*) userdata;
	short int mx,my;
	mx = LOWORD(lParam);
	my = HIWORD(lParam);
	switch(msg)
	{
	case WM_LBUTTONDOWN:
		SetCapture(hWnd);
		floating_l_down = true;
		floating_mx = mx;
		floating_my = my;
		return 0;
	case WM_LBUTTONUP:
		ReleaseCapture();
		floating_l_down = false;
		return 0;
	case WM_MOUSEMOVE:
		if(floating_l_down)
		{
			int dmx, dmy;
			dmx = mx - floating_mx;
			dmy = my - floating_my;

			bmpwnd->Move(bmpwnd->GetX()+dmx, bmpwnd->GetY()+dmy);
			bmpwnd->Redraw();
		}
		return 0;
	}
	return DefWindowProc(hWnd,msg,wParam,lParam);
}


LRESULT WINAPI WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_CREATE:
		g_hWnd = hWnd;
		ResizeBmp();
		return 0;
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	case WM_SIZE:
		switch(wParam)
		{
		case SIZE_MAXHIDE:
		case SIZE_MAXSHOW:
		case SIZE_MAXIMIZED:
		case SIZE_RESTORED:
			ResizeBmp();
		}
		return 0;
	case WM_CHAR:
		if(debug_chars) DoDebugChar(wParam);
		return 0;
	case WM_KEYDOWN:
		switch(wParam)
		{
		case VK_UP:
			key_up = true; break;
		case VK_DOWN:
			key_down = true; break;
		case VK_LEFT:
			key_left = true; break;
		case VK_RIGHT:
			key_right = true; break;
		}
		return 0;
	case WM_KEYUP:
		switch(wParam)
		{
		case VK_UP:
			key_up = false; break;
		case VK_DOWN:
			key_down = false; break;
		case VK_LEFT:
			key_left = false; break;
		case VK_RIGHT:
			key_right = false; break;	
		}
		return 0;
	case WM_LBUTTONDOWN:
		if(WonGame)
		{
			GenerateCavern();
		}
		return 0;

	case WM_PAINT:
		{
			HDC hDC;
			PAINTSTRUCT ps;
			hDC = BeginPaint(hWnd,&ps);

			Render();
			BitBlt(hDC,0,0,bmp.getWidth(),bmp.getHeight(),bmp.DC(),0,0,SRCCOPY);

			EndPaint(hWnd,&ps);

			RECT rc; // Update for the next frame :)
			GetClientRect(hWnd,&rc);
			InvalidateRect(hWnd,&rc,FALSE);
		}
		return 0;

	}
	return DefWindowProc(hWnd,msg,wParam,lParam);
}






int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR cmdline, int nsc)
{
	srand(GetTickCount());

	// Echoes seem to hurt rather than help, I wanted to change the delay time based on cavern size, but alas, my sounds are almost purely sinusoidal
	//audio.AddFilter(&echoes); 
	audio.Start();


	g_hInst = hInst;
	g_hWnd = 0;

	WNDCLASS wc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hbrBackground=0;
	wc.hCursor=LoadCursor(NULL,IDC_ARROW);
	wc.hIcon=0;
	wc.hInstance=hInst;
	wc.lpfnWndProc=&WndProc;
	wc.lpszClassName="testcls";
	wc.lpszMenuName=0;
	wc.style=CS_HREDRAW|CS_VREDRAW;
	RegisterClass(&wc);

	// Create events
	
	hRenderExitThread = CreateEvent(NULL,TRUE,FALSE,NULL);
	hRenderExitedThread = CreateEvent(NULL,TRUE,FALSE,NULL);
	hRenderCreateLevel = CreateEvent(NULL,TRUE,FALSE,NULL);
	hRenderCreatedLevel = CreateEvent(NULL,TRUE,FALSE,NULL);
	hRenderDrawLevel = CreateEvent(NULL,TRUE,FALSE,NULL);
	hRenderDrewLevel = CreateEvent(NULL,TRUE,FALSE,NULL);
	hRenderLockCavern = CreateEvent(NULL,TRUE,FALSE,NULL);
	hRenderLockedCavern = CreateEvent(NULL,TRUE,FALSE,NULL);
	hRenderUnlockCavern = CreateEvent(NULL,TRUE,FALSE,NULL);

	hMusicExitThread = CreateEvent(NULL,TRUE,FALSE,NULL);
	hMusicExitedThread = CreateEvent(NULL,TRUE,FALSE,NULL);


	if(debug_windows)
	{
		debug_backbmp = new BitmapWindow(800,10,500,500);
		debug_backbmp->SetWndProc(FloatingWndProc,debug_backbmp);
	}


	HWND hWnd = CreateWindow("testcls","Cavern Ghost",WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX,CW_USEDEFAULT,CW_USEDEFAULT,640,480,0,0,hInst,0);

	if(!hWnd)
	{
		MessageBox(0,"Error creating window","Cavern Ghost",MB_ICONERROR);
		return 1;
	}

	hRenderThread = CreateThread(NULL,0,RenderThreadProc,0,0,0);
	hMusicThread = CreateThread(NULL,0,MusicThreadProc,0,0,0);


	ShowWindow(hWnd,SW_SHOW);


	GenerateCavern();


	GetSystemTimeAsFileTime((LPFILETIME)&gamestarttime);
	lasttime = gamestarttime;


	MSG msg;
	do {
		if(PeekMessage(&msg,0,0,0,PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		else
		{
			RECT rc;
			GetClientRect(hWnd,&rc);
			InvalidateRect(hWnd,&rc,FALSE);
		}
	} while(msg.message != WM_QUIT);


	// Try to teardown the threads
	SetEvent(hRenderExitThread); // Tell the renderthread to exit
	SetEvent(hMusicExitThread); // Tell the music thread to exit
	if(WAIT_OBJECT_0!=WaitForSingleObject(hRenderExitedThread,1000)) // Give thread 1 second to exit
	{ // Otherwise kill it
		TerminateThread(hMusicThread,0);
	}
	if(WAIT_OBJECT_0!=WaitForSingleObject(hRenderExitedThread,30000)) // Give thread 30 seconds to exit
	{ // Otherwise kill it
		TerminateThread(hRenderThread,0);
	}

	return 0;
}